@@ -35,13 +35,18 @@ import tempfile
 import zipfile
 
 import xarray
+import fsspec
 import numpy as np
+import pyproj
+from pyproj.exceptions import CRSError
+
+from pygeoapi.api import DEFAULT_STORAGE_CRS
 
 from pygeoapi.provider.base import (BaseProvider,
                                     ProviderConnectionError,
                                     ProviderNoDataError,
                                     ProviderQueryError)
-from pygeoapi.util import read_data
+from pygeoapi.util import get_crs_from_uri, read_data
 
 LOGGER = logging.getLogger(__name__)
 
@@ -67,126 +72,48 @@ class XarrayProvider(BaseProvider):
                     open_func = xarray.open_mfdataset
                 else:
                     open_func = xarray.open_dataset
+            if provider_def['data'].startswith('s3://'):
+                LOGGER.debug('Data is stored in S3 bucket.')
+                if 's3' in provider_def.get('options', {}):
+                    s3_options = provider_def['options']['s3']
+                else:
+                    s3_options = {}
+                LOGGER.debug(s3_options)
+                data_to_open = fsspec.get_mapper(self.data,
+                                                 **s3_options)
+                LOGGER.debug('Completed S3 Open Function')
+            else:
+                data_to_open = self.data
 
-            self._data = open_func(self.data)
+            self._data = open_func(data_to_open)
+            self.storage_crs = self._parse_storage_crs(provider_def)
             self._coverage_properties = self._get_coverage_properties()
 
             self.axes = [self._coverage_properties['x_axis_label'],
                          self._coverage_properties['y_axis_label'],
                          self._coverage_properties['time_axis_label']]
 
-            self.fields = self._coverage_properties['fields']
+            self.get_fields()
         except Exception as err:
             LOGGER.warning(err)
             raise ProviderConnectionError(err)
 
-    def get_coverage_domainset(self, *args, **kwargs):
-        """
-        Provide coverage domainset
+    def get_fields(self):
+        if not self._fields:
+            for key, value in self._data.variables.items():
+                if len(value.shape) >= 3:
+                    LOGGER.debug('Adding variable')
+                    dtype = value.dtype
+                    if dtype.name.startswith('float'):
+                        dtype = 'number'
 
-        :returns: CIS JSON object of domainset metadata
-        """
-
-        c_props = self._coverage_properties
-        domainset = {
-            'type': 'DomainSet',
-            'generalGrid': {
-                'type': 'GeneralGridCoverage',
-                'srsName': c_props['bbox_crs'],
-                'axisLabels': [
-                    c_props['x_axis_label'],
-                    c_props['y_axis_label'],
-                    c_props['time_axis_label']
-                ],
-                'axis': [{
-                    'type': 'RegularAxis',
-                    'axisLabel': c_props['x_axis_label'],
-                    'lowerBound': c_props['bbox'][0],
-                    'upperBound': c_props['bbox'][2],
-                    'uomLabel': c_props['bbox_units'],
-                    'resolution': c_props['resx']
-                }, {
-                    'type': 'RegularAxis',
-                    'axisLabel': c_props['y_axis_label'],
-                    'lowerBound': c_props['bbox'][1],
-                    'upperBound': c_props['bbox'][3],
-                    'uomLabel': c_props['bbox_units'],
-                    'resolution': c_props['resy']
-                },
-                    {
-                        'type': 'RegularAxis',
-                        'axisLabel': c_props['time_axis_label'],
-                        'lowerBound': c_props['time_range'][0],
-                        'upperBound': c_props['time_range'][1],
-                        'uomLabel': c_props['restime'],
-                        'resolution': c_props['restime']
+                    self._fields[key] = {
+                        'type': dtype,
+                        'title': value.attrs['long_name'],
+                        'x-ogc-unit': value.attrs.get('units')
                     }
-                ],
-                'gridLimits': {
-                    'type': 'GridLimits',
-                    'srsName': 'http://www.opengis.net/def/crs/OGC/0/Index2D',
-                    'axisLabels': ['i', 'j'],
-                    'axis': [{
-                        'type': 'IndexAxis',
-                        'axisLabel': 'i',
-                        'lowerBound': 0,
-                        'upperBound': c_props['width']
-                    }, {
-                        'type': 'IndexAxis',
-                        'axisLabel': 'j',
-                        'lowerBound': 0,
-                        'upperBound': c_props['height']
-                    }]
-                }
-            },
-            '_meta': {
-                'tags': self._data.attrs
-            }
-        }
 
-        return domainset
-
-    def get_coverage_rangetype(self, *args, **kwargs):
-        """
-        Provide coverage rangetype
-
-        :returns: CIS JSON object of rangetype metadata
-        """
-
-        rangetype = {
-            'type': 'DataRecord',
-            'field': []
-        }
-
-        for name, var in self._data.variables.items():
-            LOGGER.debug(f'Determining rangetype for {name}')
-
-            desc, units = None, None
-            if len(var.shape) >= 3:
-                parameter = self._get_parameter_metadata(
-                    name, var.attrs)
-                desc = parameter['description']
-                units = parameter['unit_label']
-
-                rangetype['field'].append({
-                    'id': name,
-                    'type': 'Quantity',
-                    'name': var.attrs.get('long_name') or desc,
-                    'encodingInfo': {
-                        'dataType': f'http://www.opengis.net/def/dataType/OGC/0/{var.dtype}'  # noqa
-                    },
-                    'nodata': 'null',
-                    'uom': {
-                        'id': f'http://www.opengis.net/def/uom/UCUM/{units}',
-                        'type': 'UnitReference',
-                        'code': units
-                    },
-                    '_meta': {
-                        'tags': var.attrs
-                    }
-                })
-
-        return rangetype
+        return self._fields
 
     def query(self, properties=[], subsets={}, bbox=[], bbox_crs=4326,
               datetime_=None, format_='json', **kwargs):
@@ -211,7 +138,7 @@ class XarrayProvider(BaseProvider):
                 return read_data(self.data)
 
         if len(properties) < 1:
-            properties = self.fields
+            properties = self.fields.keys()
 
         data = self._data[[*properties]]
 
@@ -288,9 +215,9 @@ class XarrayProvider(BaseProvider):
                 _to_datetime_string(data.coords[self.time_field].values[-1])
             ],
             "driver": "xarray",
-            "height": data.dims[self.y_field],
-            "width": data.dims[self.x_field],
-            "time_steps": data.dims[self.time_field],
+            "height": data.sizes[self.y_field],
+            "width": data.sizes[self.x_field],
+            "time_steps": data.sizes[self.time_field],
             "variables": {var_name: var.attrs
                           for var_name, var in data.variables.items()}
         }
@@ -309,13 +236,13 @@ class XarrayProvider(BaseProvider):
                 fp.seek(0)
                 return fp.read()
 
-    def gen_covjson(self, metadata, data, range_type):
+    def gen_covjson(self, metadata, data, fields):
         """
         Generate coverage as CoverageJSON representation
 
         :param metadata: coverage metadata
         :param data: rasterio DatasetReader object
-        :param range_type: range type list
+        :param fields: fields
 
         :returns: dict of CoverageJSON representation
         """
@@ -324,6 +251,11 @@ class XarrayProvider(BaseProvider):
         minx, miny, maxx, maxy = metadata['bbox']
         mint, maxt = metadata['time']
 
+        selected_fields = {
+            key: value for key, value in self.fields.items()
+            if key in fields
+        }
+
         try:
             tmp_min = data.coords[self.y_field].values[0]
         except IndexError:
@@ -372,34 +304,31 @@ class XarrayProvider(BaseProvider):
             'ranges': {}
         }
 
-        for variable in range_type:
-            pm = self._get_parameter_metadata(
-                variable, self._data[variable].attrs)
-
+        for key, value in selected_fields.items():
             parameter = {
                 'type': 'Parameter',
-                'description': pm['description'],
+                'description': value['title'],
                 'unit': {
-                    'symbol': pm['unit_label']
+                    'symbol': value['x-ogc-unit']
                 },
                 'observedProperty': {
-                    'id': pm['observed_property_id'],
+                    'id': key,
                     'label': {
-                        'en': pm['observed_property_name']
+                        'en': value['title']
                     }
                 }
             }
 
-            cj['parameters'][pm['id']] = parameter
+            cj['parameters'][key] = parameter
 
         data = data.fillna(None)
         data = _convert_float32_to_float64(data)
 
         try:
-            for key in cj['parameters'].keys():
+            for key, value in selected_fields.items():
                 cj['ranges'][key] = {
                     'type': 'NdArray',
-                    'dataType': str(self._data[variable].dtype),
+                    'dataType': value['type'],
                     'axisNames': [
                         'y', 'x', self._coverage_properties['time_axis_label']
                     ],
@@ -417,19 +346,21 @@ class XarrayProvider(BaseProvider):
     def _get_coverage_properties(self):
         """
         Helper function to normalize coverage properties
+        :param provider_def: provider definition
 
         :returns: `dict` of coverage properties
         """
 
         time_var, y_var, x_var = [None, None, None]
+
         for coord in self._data.coords:
             if coord.lower() == 'time':
                 time_var = coord
                 continue
-            if self._data.coords[coord].attrs['units'] == 'degrees_north':
+            if self._data.coords[coord].attrs.get('units') == 'degrees_north':
                 y_var = coord
                 continue
-            if self._data.coords[coord].attrs['units'] == 'degrees_east':
+            if self._data.coords[coord].attrs.get('units') == 'degrees_east':
                 x_var = coord
                 continue
 
@@ -441,7 +372,7 @@ class XarrayProvider(BaseProvider):
             self.time_field = time_var
 
         # It would be preferable to use CF attributes to get width
-        # resolution etc but for now a generic approach is used to asess
+        # resolution etc but for now a generic approach is used to assess
         # all of the attributes based on lat lon vars
 
         properties = {
@@ -464,9 +395,9 @@ class XarrayProvider(BaseProvider):
             'x_axis_label': self.x_field,
             'y_axis_label': self.y_field,
             'time_axis_label': self.time_field,
-            'width': self._data.dims[self.x_field],
-            'height': self._data.dims[self.y_field],
-            'time': self._data.dims[self.time_field],
+            'width': self._data.sizes[self.x_field],
+            'height': self._data.sizes[self.y_field],
+            'time': self._data.sizes[self.time_field],
             'time_duration': self.get_time_coverage_duration(),
             'bbox_units': 'degrees',
             'resx': np.abs(self._data.coords[self.x_field].values[1]
@@ -476,13 +407,21 @@ class XarrayProvider(BaseProvider):
             'restime': self.get_time_resolution()
         }
 
-        if 'crs' in self._data.variables.keys():
-            properties['bbox_crs'] = f'http://www.opengis.net/def/crs/OGC/1.3/{self._data.crs.epsg_code}'  # noqa
+        # Update properties based on the xarray's CRS
+        epsg_code = self.storage_crs.to_epsg()
+        LOGGER.debug(f'{epsg_code}')
+        if epsg_code == 4326 or self.storage_crs == 'OGC:CRS84':
+            pass
+            LOGGER.debug('Confirmed default of WGS 84')
+        else:
+            properties['bbox_crs'] = \
+                f'https://www.opengis.net/def/crs/EPSG/0/{epsg_code}'
+            properties['inverse_flattening'] = \
+                self.storage_crs.ellipsoid.inverse_flattening
+            if self.storage_crs.is_projected:
+                properties['crs_type'] = 'ProjectedCRS'
 
-            properties['inverse_flattening'] = self._data.crs.\
-                inverse_flattening
-
-            properties['crs_type'] = 'ProjectedCRS'
+        LOGGER.debug(f'properties: {properties}')
 
         properties['axes'] = [
             properties['x_axis_label'],
@@ -490,9 +429,6 @@ class XarrayProvider(BaseProvider):
             properties['time_axis_label']
         ]
 
-        properties['fields'] = [name for name in self._data.variables
-                                if len(self._data.variables[name].shape) >= 3]
-
         return properties
 
     @staticmethod
@@ -551,6 +487,71 @@ class XarrayProvider(BaseProvider):
 
         return ', '.join(times)
 
+    def _parse_grid_mapping(self):
+        """
+        Identifies grid_mapping.
+
+        :returns: name of xarray data variable that contains CRS information.
+        """
+        LOGGER.debug('Parsing grid mapping...')
+        spatiotemporal_dims = (self.time_field, self.y_field, self.x_field)
+        LOGGER.debug(spatiotemporal_dims)
+        grid_mapping_name = None
+        for var_name, var in self._data.variables.items():
+            if all(dim in var.dims for dim in spatiotemporal_dims):
+                try:
+                    grid_mapping_name = self._data[var_name].attrs['grid_mapping']  # noqa 
+                    LOGGER.debug(f'Grid mapping: {grid_mapping_name}')
+                except KeyError as err:
+                    LOGGER.debug(err)
+                    LOGGER.debug('No grid mapping information found.')
+        return grid_mapping_name
+
+    def _parse_storage_crs(
+        self,
+        provider_def: dict
+    ) -> pyproj.CRS:
+        """
+        Parse the storage CRS from an xarray dataset.
+
+        :param provider_def: provider definition
+
+        :returns: `pyproj.CRS` instance parsed from dataset
+        """
+        storage_crs = None
+
+        try:
+            storage_crs = provider_def['storage_crs']
+            crs_function = pyproj.CRS.from_user_input
+        except KeyError as err:
+            LOGGER.debug(err)
+            LOGGER.debug('No storage_crs found. Attempting to parse the CRS.')
+
+        if storage_crs is None:
+            grid_mapping = self._parse_grid_mapping()
+            if grid_mapping is not None:
+                storage_crs = self._data[grid_mapping].attrs
+                crs_function = pyproj.CRS.from_cf
+            elif 'crs' in self._data.variables.keys():
+                storage_crs = self._data['crs'].attrs
+                crs_function = pyproj.CRS.from_dict
+            else:
+                storage_crs = DEFAULT_STORAGE_CRS
+                crs_function = get_crs_from_uri
+                LOGGER.debug('Failed to parse dataset CRS. Assuming WGS84.')
+
+        LOGGER.debug(f'Parsing CRS {storage_crs} with {crs_function}')
+        try:
+            crs = crs_function(storage_crs)
+        except CRSError as err:
+            LOGGER.debug(f'Unable to parse projection with pyproj: {err}')
+            LOGGER.debug('Assuming default WGS84.')
+            crs = get_crs_from_uri(DEFAULT_STORAGE_CRS)
+
+        LOGGER.debug(crs)
+
+        return crs
+
 
 def _to_datetime_string(datetime_obj):
     """
